/**
 * DREAM
 * Copyright (C) 2003 2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s):
 */

package org.objectweb.dream.control.activity.task.thread;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.control.activity.scheduler.Scheduler;
import org.objectweb.dream.control.activity.scheduler.StoppedSchedulerException;
import org.objectweb.dream.control.activity.task.Task;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Abstract class representing a thread task.
 */
public abstract class AbstractThreadTask extends AbstractComponent
    implements
      Task
{

  /** The scheduler to which the thread is bound. */
  protected Scheduler scheduler = null;

  /**
   * Called by {@link #execute(Object) }method.
   * 
   * @return <code>false</code> if the {@link #execute(Object) }method must
   *         return.
   */
  protected abstract boolean isExecuting();

  /**
   * Called by {@link #execute(Object)}method. Sets its execution state.
   * 
   * @param b <code>false</code> if the {@link #execute(Object) }method must
   *          return.
   */
  protected abstract void setExecuting(boolean b);

  // ---------------------------------------------------------------------------
  // Implementation of the Task interface
  // ---------------------------------------------------------------------------

  /**
   * @see Task#execute(Object)
   */
  public Object execute(Object hints) throws InterruptedException
  {
    try
    {
      while (isExecuting())
      {
        logger.log(BasicLevel.DEBUG, "Call to schedule");
        if (((Integer) scheduler.schedule(null)).intValue() == -1)
        {
          setExecuting(false);
        }
      }
    }
    catch (StoppedSchedulerException e1)
    {
      logger.log(BasicLevel.DEBUG, "Scheduler stopped");
    }
    catch (InterruptedException e)
    {
      logger.log(BasicLevel.DEBUG, "Interrupted");
    }
    catch (NullPointerException e)
    {
      // In case where the unbind has been done.
    }
    logger.log(BasicLevel.DEBUG, "End of the execute method");
    return null;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(Scheduler.ITF_NAME))
    {
      scheduler = (Scheduler) serverItf;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#unbindFc(String)
   */
  public synchronized void unbindFc(String clientItfName)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.unbindFc(clientItfName);
    if (clientItfName.equals(Scheduler.ITF_NAME))
    {
      scheduler = null;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public synchronized String[] listFc()
  {
    return new String[]{Scheduler.ITF_NAME};
  }

}