/**
 * DREAM
 * Copyright (C) 2003 2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Vivien Quema
 */

package org.objectweb.dream.message.manager;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.message.AbstractNonExtensibleMessage;
import org.objectweb.dream.message.ChunkType;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageType;
import org.objectweb.dream.util.EmptyStringArray;
import org.objectweb.dream.util.Error;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class provides an implementation of message managers for non extensible
 * messages. It pools message instances.
 * <p>
 * The class of messages to be managed is given using the
 * <code>NonExtensibleMessageManagerAttributeController</code>.
 */
public class PoolNonExtensibleMessageManagerImpl extends AbstractComponent
    implements
      MessageManager,
      PoolNonExtensibleMessageManagerAttributeController
{

  private Message pool[];
  private int     elementCount = 0;

  private int     capacity;

  // the id of this message manager.
  private short   id;

  // The name of the class of managed messages
  private String  className    = null;
  // The class of messages managed by this message manager
  private Class   messageClass;

  // ---------------------------------------------------------------------------
  // Implementation of the MessageManager interface
  // ---------------------------------------------------------------------------

  /**
   * @see MessageManager#createMessage(MessageType)
   */
  public synchronized Message createMessage(MessageType type)
      throws UnknownChunkTypeError
  {
    // Ignore parameter, always return an ebs message
    if (elementCount == 0)
    {
      logger.log(BasicLevel.DEBUG, "Create new message");
      Message m = null;
      try
      {
        m = (Message) messageClass.newInstance();
      }
      catch (Exception e)
      {
        // Should not happen since it has been tested in the setClassName method
        Error.bug(logger, e);
      }
      m.setMessageManagerId(id);
      ((AbstractNonExtensibleMessage) m).incrementReferenceCounter();
      return m;
    }
    elementCount -= 1;
    Message message = pool[elementCount];
    pool[elementCount] = null;
    ((AbstractNonExtensibleMessage) message).incrementReferenceCounter();
    return message;
  }

  /**
   * @see MessageManager#deleteMessage(Message)
   */
  public synchronized void deleteMessage(Message message)
  {
    if (message != null)
    {
      if (((AbstractNonExtensibleMessage) message).decrementReferenceCounter())
      {
        message.recycle();
        if (elementCount == pool.length || message == null)
        {
          return;
        }
        pool[elementCount] = message;
        elementCount += 1;
      }
    }
  }

  /**
   * @see MessageManager#duplicateMessage(Message, boolean)
   */
  public Message duplicateMessage(Message message, boolean clone)
  {
    if (clone)
    {
      Message newInstace = createMessage(null);
      message.transfertChunkStates(newInstace);
      return newInstace;
    }
    ((AbstractNonExtensibleMessage) message).incrementReferenceCounter();
    return message;
  }

  /**
   * @see MessageManager#createChunk(ChunkType)
   */
  public Object createChunk(ChunkType type) throws UnknownChunkTypeError
  {
    Error
        .error("This message manager handle only inextensible message", logger);
    return null;
  }

  /**
   * @see MessageManager#deleteChunk(Object)
   */
  public void deleteChunk(Object chunk)
  {
    Error
        .error("This message manager handle only inextensible message", logger);
  }

  /**
   * @see MessageManager#duplicateChunk(Object, boolean)
   */
  public Object duplicateChunk(Object chunk, boolean clone)
  {
    Error
        .error("This message manager handle only inextensible message", logger);
    return null;
  }

  /**
   * @see MessageManager#getMessageManagerId()
   */
  public short getMessageManagerId()
  {
    return id;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the MessageManagerAttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see MessageManagerAttributeController#getId()
   */
  public short getId()
  {
    return id;
  }

  /**
   * @see MessageManagerAttributeController#setId(short)
   */
  public void setId(short id)
  {
    this.id = id;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the PoolMessageManagerAttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see PoolMessageManagerAttributeController#getCapacity()
   */
  public int getCapacity()
  {
    return capacity;
  }

  /**
   * @see PoolMessageManagerAttributeController#setCapacity(int)
   */
  public void setCapacity(int capacity)
  {
    // TODO try to re-use previous instance
    this.capacity = capacity;
    pool = new Message[capacity];
    elementCount = 0;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the NonExtensibleMessageManagerAttributeController
  // interface
  // ---------------------------------------------------------------------------

  /**
   * @see NonExtensibleMessageManagerAttributeController#getMessageClassName()
   */
  public String getMessageClassName()
  {
    return className;
  }

  /**
   * @see NonExtensibleMessageManagerAttributeController#setMessageClassName(String)
   */
  public void setMessageClassName(String name)
  {
    if (className != null)
    {
      logger.log(BasicLevel.DEBUG,
          "This message manager already manages messaeges of class "
              + className);
      new IllegalArgumentException(
          "This message manager already manages messaeges of class "
              + className);
    }
    this.className = name;
    try
    {
      messageClass = Class.forName(name);
      // Test that the class is correct
      messageClass.newInstance();
    }
    catch (Exception e)
    {
      logger.log(BasicLevel.DEBUG, "Invalid class " + className, e);
      new IllegalArgumentException("Invalid class " + className);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------
  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return EmptyStringArray.EMPTY_STRING_ARRAY;
  }

}