/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.util.monolog.wrapper.javaLog;

import org.objectweb.util.monolog.file.api.Pattern;
import org.objectweb.util.monolog.wrapper.common.AbstractFactory;

import java.util.logging.Formatter;
import java.util.logging.LogRecord;
import java.util.StringTokenizer;
import java.util.ArrayList;
import java.util.List;
import java.io.StringWriter;
import java.io.PrintWriter;

/**
 *
 * @author S.Chassande-Barrioz
 */
public class MonologFormatter extends Formatter {

    private final static int PATTERN_ID_LEVEL = -100;
    private final static int PATTERN_ID_TOPIC = -200;
    private final static int PATTERN_ID_DATE = -300;
    private final static int PATTERN_ID_THREAD = -400;
    private final static int PATTERN_ID_MESSAGE = -500;
    private final static int PATTERN_ID_METHOD = -600;
    private final static int PATTERN_ID_OBJECT = -700;
    private final static int PATTERN_ID_LINE_NUMBER = -800;
    private final static int PATTERN_ID_NEW_LINE = -900;
    private final static int PATTERN_ID_INTERVAL = 100;

    private final static String TOKENS =
            "{}"
            + Pattern.LEVEL
            + Pattern.TOPIC
            + Pattern.DATE
            + Pattern.THREAD
            + Pattern.MESSAGE
            + Pattern.METHOD
            + Pattern.OBJECT
            + Pattern.LINE_NUMBER
            + Pattern.PREFIX
            + Pattern.NEW_LINE;

    String strPattern;
    int[] pattern;
    String[] strings;

    public MonologFormatter() {
    }

    public MonologFormatter(String strPattern) {
        setPattern(strPattern);
    }


    public String getPattern() {
        return strPattern;
    }

    public void setPattern(String p) {
        debug("Pattern=" + p);
        this.strPattern = p;
        if (strPattern == null) {
            pattern = new int[0];
        } else {
            StringTokenizer st = new StringTokenizer(p, TOKENS, true);
            ArrayList sections = new ArrayList();
            boolean isPrefix = false;
            boolean isObject = false;
            boolean isInSubObject = false;
            int subObjectNumber = 1;
            while (st.hasMoreElements()) {
                String token = st.nextToken();
                debug("token=<" + token + ">");
                if (isObject && token.equals("{")) {
                    isInSubObject = true;
                    isObject = false;
                }
                if (token.length() == 1) {
                    char c = token.charAt(0);
                    switch (c) {
                    case '{':
                        if (!isInSubObject) {
                            addSection(sections, token);
                        }
                        break;
                    case '}':
                        if (isInSubObject) {
                            int old = ((Integer) sections.get(sections.size() - 1)).intValue();
                            sections.set(sections.size() - 1,
                                    new Integer(old - subObjectNumber));
                            isInSubObject = false;
                        } else {
                            addSection(sections, token);
                        }
                        break;
                    case Pattern.PREFIX:
                        if (isPrefix) {
                            sections.add(String.valueOf(Pattern.PREFIX));
                        }
                        isPrefix = !isPrefix;
                        break;
                    case Pattern.LEVEL:
                        isPrefix = treatPattern(
                                sections, token, PATTERN_ID_LEVEL, isPrefix);
                        break;
                    case Pattern.TOPIC:
                        isPrefix = treatPattern(
                                sections, token, PATTERN_ID_TOPIC, isPrefix);
                        break;
                    case Pattern.DATE:
                        isPrefix = treatPattern(
                                sections, token, PATTERN_ID_DATE, isPrefix);
                        break;
                    case Pattern.THREAD:
                        isPrefix = treatPattern(
                                sections, token, PATTERN_ID_THREAD, isPrefix);
                        break;
                    case Pattern.MESSAGE:
                        isPrefix = treatPattern(
                                sections, token, PATTERN_ID_MESSAGE, isPrefix);
                        break;
                    case Pattern.METHOD:
                        isPrefix = treatPattern(
                                sections, token, PATTERN_ID_METHOD, isPrefix);
                        break;
                    case Pattern.OBJECT:
                        isPrefix = treatPattern(
                                sections, token, PATTERN_ID_OBJECT, isPrefix);
                        isObject = true;
                        break;
                    case Pattern.LINE_NUMBER:
                        isPrefix = treatPattern(
                                sections, token, PATTERN_ID_LINE_NUMBER, isPrefix);
                        break;
                    case Pattern.NEW_LINE:
                        isPrefix = treatPattern(
                                sections, token, PATTERN_ID_NEW_LINE, isPrefix);
                        break;
                    default:
                        if (isInSubObject) {
                            subObjectNumber = Integer.parseInt(token);
                        } else {
                            addSection(sections, token);
                        }
                        break;
                    }
                } else if (isObject) {
                    //Ignore
                } else if (isInSubObject) {
                    subObjectNumber = Integer.parseInt(token);
                } else {
                    addSection(sections, token);
                }
            }
            pattern = new int[sections.size()];
            debug("building pattern array...");
            debug("nb of pattern:" + pattern.length);
            ArrayList stringList = new ArrayList(sections.size());
            int cpt = 0;
            for (int i = 0; i < pattern.length; i++) {
                Object o = sections.get(i);
                if (o instanceof String) {
                    debug("add current pattern into strings: [" + cpt + ", " + o + "]");
                    stringList.add(o);
                    pattern[i] = cpt;
                    cpt++;
                } else if (o instanceof Integer) {
                    debug("add current pattern as negative number:" + o);
                    pattern[i] = ((Integer) o).intValue();
                }
            }
            strings = (String[]) stringList.toArray(new String[cpt]);
            debug("nb of string:" + strings.length);
        }
    }

    private boolean treatPattern(List sections,
                                 String token,
                                 int tokenId,
                                 boolean isPrefix) {
        debug("treatPttern(" + tokenId + "):"
                + " isPrefix=" + isPrefix
                + " token=" + token
                + " sections=" + sections
        );
        if (isPrefix) {
            sections.add(new Integer(tokenId));
            return false;
        } else {
            addSection(sections, token);
            return isPrefix;
        }
    }

    private void addSection(List sections, String s) {
        int size = sections.size();
        if (size == 0) {
            debug("addSection(" + s + ", " + sections + "): first elem");
            sections.add(s);
        } else {
            Object last = sections.get(size - 1);
            if (last instanceof String) {
                sections.set(size - 1, last + s);
                debug("addSection(" + s + ", " + sections + "): concat: " + sections.get(size - 1));
            } else {
                debug("addSection(" + s + ", " + sections + "): new elem");
                sections.add(s);
            }
        }
    }

    /**
     * Format the given log record and return the formatted string.
     * <p>
     * The resulting formatted String will normally include a
     * localized and formated version of the LogRecord's message field.
     * The Formatter.formatMessage convenience method can (optionally)
     * be used to localize and format the message field.
     *
     * @param record the log record to be formatted.
     * @return the formatted log record
     */
    public String format(LogRecord record) {
        StringBuffer sb = new StringBuffer();
        String[] ctx = null;
        for (int i = 0; i < pattern.length; i++) {
            int p = pattern[i];
            debug("format: pattern=" + p);
            switch (p) {
            case PATTERN_ID_LEVEL:
                sb.append(record.getLevel().getName());
                break;
            case PATTERN_ID_TOPIC:
                sb.append(record.getLoggerName());
                break;
            case PATTERN_ID_DATE:
                sb.append(record.getMillis());
                break;
            case PATTERN_ID_THREAD:
                sb.append(record.getThreadID());
                break;
            case PATTERN_ID_MESSAGE:
                sb.append(record.getMessage());
                break;
            case PATTERN_ID_METHOD:
                if (ctx == null) {
                    ctx = getContext();
                }
                sb.append(ctx[1]);
                break;
            case PATTERN_ID_OBJECT:
                if (ctx == null) {
                    ctx = getContext();
                }
                sb.append(ctx[0]);
                break;
            case PATTERN_ID_LINE_NUMBER:
                if (ctx == null) {
                    ctx = getContext();
                }
                sb.append(ctx[2]);
                break;
            case PATTERN_ID_NEW_LINE:
                sb.append("\n");
                break;
            default:
                if (p < 0) {
                    if (ctx == null) {
                        ctx = getContext();
                    }
                    if (p > (PATTERN_ID_OBJECT - PATTERN_ID_INTERVAL)
                            && p<PATTERN_ID_OBJECT) {
                        p = PATTERN_ID_OBJECT - p;
                        String res = ctx[0];
                        if (p == 1) {// optimize the usual case
                            int idx = res.lastIndexOf('.');
                            if (idx != -1) {
                                res = res.substring(idx + 1);
                            }
                        } else if (p == 0) {
                            //Nothing to do
                        } else {
                            //TODO: support %O{n} with n > 1
                        }
                        sb.append(res);
                    }
                } else if (p >= strings.length) {
                    System.err.println("ERROR: String identifier unknown: " + p);
                } else {
                    sb.append(strings[p]);
                }
            }
        }
        if (record.getThrown() != null) {
            StringWriter sw = new StringWriter();
            record.getThrown().printStackTrace(new PrintWriter(sw));
            sb.append(sw.getBuffer());
        }
        return sb.toString();
    }

    private static void debug(String m) {
        AbstractFactory.debug(m);
    }

    public static String[] getContext() {
        int removeTopStack = 10;

        Throwable t = new Throwable().fillInStackTrace();
        StringWriter sw = new StringWriter();
        t.printStackTrace(new PrintWriter(sw));
        String m = sw.getBuffer().toString();
        int fin = 0;
        int deb = -1;
        for (int i = 0; i < removeTopStack; i++) {
            deb = m.indexOf("\n", deb + 1);
        }
        deb = m.indexOf("at ", deb);
        fin = m.indexOf("\n", deb);
        m = m.substring(deb + 3, fin);

        //m = %C.%M(Toto.java:%L)
        deb = m.indexOf("(");
        fin = m.indexOf(":");
        String[] res = new String[3];
        res[2] = (fin == -1 ? "unknown" : m.substring(fin + 1, m.length() - 1));
        m = m.substring(0, deb);

        //m = %C.%M
        fin = m.indexOf('.');
        res[0] = m.substring(0, fin);
        res[1] = m.substring(fin + 1);
        return res;
    }

    /**
     * This method permits to format messages. More exatcly this method find
     * the class name and the method name where the log call was done.
     * In order to find the right class name and method name, this method is
     * parametrable either the number of call done in this logger.
     *
     public static String[] getContext(String msg, int removeTopStack) {
     Throwable t = new Throwable().fillInStackTrace();
     StringWriter sw = new StringWriter();
     t.printStackTrace(new PrintWriter(sw));
     String m = sw.getBuffer().toString();

     int deb = -1;

     int fin = 0;

     // take the right line
     deb = -1;
     for (int i = 0; i < (removeTopStack + 1); i++) {
     deb = m.indexOf("\n", deb + 1);
     }

     deb = m.indexOf("at ", deb);
     fin = m.indexOf("\n", deb);
     m = m.substring(deb + 3, fin);

     // remove param
     deb = m.indexOf("(");
     fin = m.indexOf(":");
     m = m.substring(0, deb + 1) + m.substring(fin + 1, m.length());

     // remove package name
     deb = m.indexOf("(");
     int c1 = 0;
     int c2 = 0;
     int current = m.indexOf(".");
     while (current != -1 && current < deb) {
     c1 = c2;
     c2 = current;
     current = m.indexOf(".", current + 1);
     }
     m = m.substring(c1 + 1, m.length());

     return m + ": " + msg;
     }
     /* */

}
